from functools import partial
import os
from typing import TYPE_CHECKING

import numpy as np
import skimage.io
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QAbstractItemView,
    QFileDialog,
    QHBoxLayout,
    QLineEdit,
    QListWidget,
    QMessageBox,
    QPushButton,
    QVBoxLayout,
    QWidget,
)
import yaml

from ._utils import find_missing_number, xywh2xyxy, save_text

if TYPE_CHECKING:
    import napari


class BboxQWidget(QWidget):
    def __init__(self, napari_viewer):
        super().__init__()
        self.viewer = napari_viewer
        self.initUI()
        self.initVariables()
        self.initLayers()

    def initUI(self):
        # Create button for opening a file
        self.open_file_button = QPushButton("Open File", self)
        self.open_file_button.clicked.connect(self.openFile)

        # Create button for opening a directory
        self.open_dir_button = QPushButton("Open Directory", self)
        self.open_dir_button.clicked.connect(self.openDirectory)

        # Create a list widget for displaying the list of opened files
        self.listWidget = QListWidget()
        self.listWidget.itemClicked.connect(self.open_image)

        # Create button for clear the list of opened files
        self.clear_button = QPushButton("Clear list of opened files", self)
        self.clear_button.clicked.connect(self.listWidget.clear)

        # Create a list widget for displaying the list of classes
        self.classlistWidget = QListWidget()
        self.classlistWidget.setSelectionMode(QAbstractItemView.SingleSelection)
        self.classlistWidget.itemClicked.connect(self.class_clicked)

        # Create text box for entering the class names
        self.class_textbox = QLineEdit()
        self.class_textbox.setPlaceholderText("Enter class name")

        # Create button for adding class to classlist
        self.add_class_button = QPushButton("Add class", self)
        self.add_class_button.clicked.connect(self.add_class)

        # Create button for deleting class from classlist
        self.del_class_button = QPushButton("Delete selected class", self)
        self.del_class_button.clicked.connect(self.del_class)

        # Create button for saving the bounding box annotations
        self.save_button = QPushButton("Save Annotations", self)
        self.save_button.clicked.connect(self.saveAnnotations)

        # Set the layout
        layout = QVBoxLayout()
        hbox = QHBoxLayout()
        hbox.addWidget(self.open_file_button)
        hbox.addWidget(self.open_dir_button)
        layout.addLayout(hbox)
        layout.addWidget(self.listWidget)
        layout.addWidget(self.clear_button)
        layout.addWidget(self.classlistWidget)
        layout.addWidget(self.class_textbox)
        layout.addWidget(self.add_class_button)
        layout.addWidget(self.del_class_button)
        layout.addWidget(self.save_button)
        self.setLayout(layout)

    def initVariables(self):
        """Initializes the variables."""
        self.features = {"class": []}
        self.text = {
            "string": "{class}",
            "anchor": "upper_left",
            "translation": [0, 0],
            "size": 10,
            "color": "green",
        }
        self.numbers = []
        self.current_class_number = 0

    def initLayers(self):
        """Initializes the image and shapes layers in the napari viewer."""
        self.viewer.add_image(np.zeros((10, 10)), name="image_layer")
        self.viewer.add_shapes(name="bbox_layer", features=self.features, text=self.text)
        # self.viewer.layers["bbox_layer"].mouse_drag_callbacks.append(self.add_size)

    def class_clicked(self):
        self.set_default_class()
        # change class if shapes are selected
        shapes_layer = self.viewer.layers["bbox_layer"]
        selected_item = self.classlistWidget.selectedItems()[0]
        if not selected_item:
            return
        idxs = list(shapes_layer.selected_data)
        if len(idxs) != 0:
            class_name = selected_item.text()
            shapes_layer.features.loc[idxs, "class"] = class_name
            shapes_layer.refresh_text()

    def set_default_class(self):
        """Sets the default class for the shapes layer."""
        shapes_layer = self.viewer.layers["bbox_layer"]
        selected_item = self.classlistWidget.selectedItems()[0]
        if not selected_item:
            return
        print(shapes_layer.feature_defaults)
        shapes_layer.feature_defaults["class"] = selected_item.text()
        print(type(shapes_layer.selected_data))
        print(self.text)

    def add_class(self):
        """Adds the text in the class_textbox to the classlistWidget."""
        class_name = self.class_textbox.text()
        if class_name:
            exist_class_names = [self.classlistWidget.item(i).text() for i in range(self.classlistWidget.count())]
            if len(exist_class_names) == 0:
                self.current_class_number = 0
            else:
                if class_name in exist_class_names:
                    print("Class already exists")
                    return
                self.numbers = [int(name.split(":")[0]) for name in exist_class_names]
                self.current_class_number = find_missing_number(self.numbers)
                print("current class number:", self.current_class_number)
                if self.current_class_number != len(exist_class_names):
                    self.popup("numbering")
            class_name = f"{self.current_class_number}: {class_name}"
            self.classlistWidget.addItem(class_name)
            self.sort_classlist()
            self.class_textbox.clear()

    def popup(self, message_type=None):
        if message_type == "None":
            return
        popup = QMessageBox(self)
        if message_type == "numbering":
            popup.setWindowTitle("Numbering")
            popup.setText(
                f"Insert new class as item number {self.current_class_number} or append as the next highest number?"
            )
            popup.setStandardButtons(QMessageBox.Cancel | QMessageBox.No | QMessageBox.Yes)
            popup.button(QMessageBox.No).setText("Append")
            popup.button(QMessageBox.Yes).setText("Insert")
            popup.buttonClicked.connect(self.on_popup_button_clicked_numbering)
        elif message_type == "renumbering":
            popup.setWindowTitle("Renumbering")
            popup.setText(
                "Do you want to renumber the classes? If you click 'Yes', the classes will be renumbered from 1. *Note that this will NOT change the class numbers in the existing annotations.*"
            )
            popup.setStandardButtons(QMessageBox.Cancel | QMessageBox.No | QMessageBox.Yes)
            popup.buttonClicked.connect(self.on_popup_button_clicked_renumbering)
        popup.exec_()

    def on_popup_button_clicked_numbering(self, button):
        if button.text() == "Insert":
            return
        elif button.text() == "Append":
            self.current_class_number = max(self.numbers) + 1
        elif button.text() == "Cancel":
            self.current_class_number = 0
        else:
            self.current_class_number = 0

    def on_popup_button_clicked_renumbering(self, button):
        print(button.text())
        if button.text() == "Cancel":
            return
        else:
            selected_item = self.classlistWidget.selectedItems()[0]
            self.classlistWidget.takeItem(self.classlistWidget.row(selected_item))
            if button.text() == "&Yes":
                self.sort_classlist(renumber=True)
            else:
                pass

    def sort_classlist(self, renumber=False):
        items_text = [self.classlistWidget.item(i).text() for i in range(self.classlistWidget.count())]

        def extract_number(item_text):
            return int(item_text.split(":")[0].strip())

        sorted_items_text = sorted(items_text, key=extract_number)

        if renumber:
            renumbered_items_text = []
            for idx, item_text in enumerate(sorted_items_text):
                _, text = item_text.split(":", 1)
                renumbered_items_text.append(f"{idx}: {text.strip()}")
            sorted_items_text = renumbered_items_text

        self.classlistWidget.clear()
        for item_text in sorted_items_text:
            self.classlistWidget.addItem(item_text)

    def del_class(self):
        """Deletes the selected class from the classlistWidget and the features dictionary."""
        if not self.classlistWidget.selectedItems():
            return
        self.popup("renumbering")

    def openFile(self):
        fname = QFileDialog.getOpenFileName(self, "Open file", "/")
        if fname[0]:
            self.listWidget.addItem(fname[0])
            item = self.listWidget.findItems(fname[0], Qt.MatchExactly)[0]
            self.listWidget.setCurrentItem(item)
            self.open_image(item)

    def openDirectory(self):
        dname = QFileDialog.getExistingDirectory(self, "Open directory", "/")
        if dname:
            files = os.listdir(dname)
            image_files = sorted([f for f in files if f.endswith((".png", ".jpg", ".jpeg", ".tif", ".tiff"))])
            for image_file in image_files:
                self.listWidget.addItem(os.path.join(dname, image_file))

    def popup_load_class(self, class_data_from_yaml):
        popup = QMessageBox(self)
        popup.setWindowTitle("Load Classlist")
        popup.setText("Do you want to load and overwrite the existing classlist?")
        popup.setStandardButtons(QMessageBox.Cancel | QMessageBox.Yes)
        popup.buttonClicked.connect(partial(self.on_popup_button_clicked_load_class, class_data_from_yaml))
        popup.exec_()

    def on_popup_button_clicked_load_class(self, class_data_from_yaml, clicked_button):
        if clicked_button.text() == "Cancel":
            return
        else:
            self.classlistWidget.clear()
            for class_id, class_name in class_data_from_yaml["names"].items():
                self.classlistWidget.addItem(f"{class_id}: {class_name}")
            self.sort_classlist()

    def open_image(self, item):
        """Opens an image and updates the image layer in the napari viewer."""
        image_file = item.text()
        image = skimage.io.imread(image_file)
        rgb = image.shape[-1] in (3, 4)
        if rgb:
            image_height, image_width, _ = image.shape[-3:]
        else:
            image_height, image_width = image.shape[-2:]
        image_layer = self.viewer.layers["image_layer"]
        image_layer.rgb = rgb
        image_layer.data = image
        image_layer.reset_contrast_limits()

        classes = []

        class_file = os.path.dirname(image_file) + "/class.yaml"
        if os.path.isfile(class_file):
            with open(class_file, "r") as file:
                class_data_from_yaml = yaml.safe_load(file)
            print(class_data_from_yaml)
            if self.classlistWidget.count() != 0:
                items_text = [self.classlistWidget.item(i).text() for i in range(self.classlistWidget.count())]
                items_dict = {
                    int(item_text.split(":")[0].strip()): item_text.split(":")[1].strip() for item_text in items_text
                }
                class_data = {"names": items_dict}
                if class_data_from_yaml != class_data:
                    self.popup_load_class(class_data_from_yaml)
                else:
                    pass
            else:
                for class_id, class_name in class_data_from_yaml["names"].items():
                    self.classlistWidget.addItem(f"{class_id}: {class_name}")
            self.sort_classlist()
        items_text = [self.classlistWidget.item(i).text() for i in range(self.classlistWidget.count())]
        self.numbers = [int(name.split(":")[0]) for name in items_text]

        items_dict_with_no = {
            item_text.split(":")[0].strip(): item_text.split(":")[1].strip() for item_text in items_text
        }

        txt_file = os.path.splitext(image_file)[0] + ".txt"
        if os.path.exists(txt_file):
            with open(txt_file, "r") as f:
                lines = f.readlines()
                shapes_data = []
                for line in lines:
                    class_id, x_center, y_center, width, height = map(float, line.strip().split())
                    x_min, y_min, x_max, y_max = xywh2xyxy(
                        [x_center, y_center, width, height], scale=(image_width, image_height)
                    )
                    shapes_data.append([[y_min, x_min], [y_min, x_max], [y_max, x_max], [y_max, x_min]])
                    # TODO: Add function to convert class_id to class name
                    if str(int(class_id)) in items_dict_with_no.keys():
                        classes.append(str(int(class_id)) + ": " + items_dict_with_no[str(int(class_id))])
                    else:
                        self.classlistWidget.addItem(str(int(class_id)) + ": ")
                        items_dict_with_no[str(int(class_id))] = ""
                        self.numbers.append(int(class_id))
                        classes.append(str(int(class_id)) + ": ")
                shapes_layer = self.viewer.layers["bbox_layer"]
                shapes_layer.data = shapes_data
                shapes_layer.features["class"] = classes
                self.sort_classlist()
                shapes_layer.refresh_text()
                print(shapes_layer.features)
                print(shapes_layer.text)
        else:
            shapes_layer = self.viewer.layers["bbox_layer"]
            shapes_layer.data = []

        self.viewer.reset_view()

    def saveAnnotations(self):
        """Saves the bounding box annotations in the shapes layer in YOLO format."""

        # Get the current image file name and the corresponding annotation file name
        current_image_file = self.listWidget.currentItem().text()
        annotation_file = os.path.splitext(current_image_file)[0] + ".txt"

        shapes_layer = self.viewer.layers["bbox_layer"]
        image_layer = self.viewer.layers["image_layer"]
        if image_layer.rgb:
            image_height, image_width, _ = image_layer.data.shape[-3:]
        else:
            image_height, image_width = image_layer.data.shape[-2:]
        shapes_data = shapes_layer.data

        annotations = []

        items_text = [self.classlistWidget.item(i).text() for i in range(self.classlistWidget.count())]
        items_dict = {int(item_text.split(":")[0].strip()): item_text.split(":")[1].strip() for item_text in items_text}
        class_data = {"names": items_dict}
        class_file = os.path.join(os.path.dirname(annotation_file), "class.yaml")

        # For each shape (rectangle)
        for i, shape_data in enumerate(shapes_data):
            # Calculate the center, width, and height of the shape
            y_min, x_min = map(int, shape_data[0])
            y_max, x_max = map(int, shape_data[2])

            # Clip the coordinates to the image boundaries
            y_min = np.clip(y_min, 0, image_height - 1)
            y_max = np.clip(y_max, 0, image_height - 1)
            x_min = np.clip(x_min, 0, image_width - 1)
            x_max = np.clip(x_max, 0, image_width - 1)

            x_center = ((x_max + x_min) / 2) / image_width
            y_center = ((y_max + y_min) / 2) / image_height
            width = (x_max - x_min) / image_width
            height = (y_max - y_min) / image_height

            # Append the annotation to the list
            # class_name = shapes_layer.features["class"][i].split(":")[1].strip()
            # class_id = list(items_dict.keys())[list(items_dict.values()).index(class_name)]
            class_id = shapes_layer.features["class"][i].split(":")[0].strip()
            annotations.append(f"{class_id} {x_center} {y_center} {width} {height}")

        # Join all the annotations into a string
        annotations_str = "\n".join(annotations)

        self.check_file(annotation_file, annotations_str, file_type="annotations")

        if not os.path.isfile(class_file):
            self.check_file(class_file, class_data, file_type="classlist")
        with open(class_file, "r") as file:
            prev_items_dict = yaml.safe_load(file)
        if prev_items_dict != class_data:
            self.check_file(class_file, class_data, file_type="classlist")

    def check_file(self, filepath, file_str, file_type="annotations"):
        popup = QMessageBox(self)
        if file_type == "annotations":
            popup.setWindowTitle("Save Annotations")
        elif file_type == "classlist":
            popup.setWindowTitle("Save Classlist")
        else:
            popup.setWindowTitle("Save File")

        if os.path.isfile(filepath):
            with open(filepath, "r") as f:
                if f.read() == file_str:
                    self.show_saved_popup(popup, filepath, file_str, file_type)
                else:
                    if file_type == "annotations":
                        popup.setText("Do you want to overwrite the existing annotations?")
                    elif file_type == "classlist":
                        popup.setText("Do you want to overwrite the existing classlist?")
                    else:
                        popup.setText("Do you want to overwrite the existing file?")
                    popup.setStandardButtons(QMessageBox.Cancel | QMessageBox.Yes)
                    popup.button(QMessageBox.Yes).setText("Overwrite")
                    popup.buttonClicked.connect(
                        partial(self.on_popup_button_clicked_save, filepath, file_str, file_type)
                    )
                    popup.exec_()
        else:
            self.show_saved_popup(popup, filepath, file_str, file_type)

    def on_popup_button_clicked_save(self, filepath, file_str, file_type, clicked_button):
        if clicked_button.text() == "Overwrite":
            save_text(filepath, file_str, file_type)

    def show_saved_popup(self, popup, filepath, file_str, file_type):
        save_text(filepath, file_str, file_type)
        popup.setText(f"{file_type} saved")
        popup.setStandardButtons(QMessageBox.Close)
        popup.exec_()
